/*oftware are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2012 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.tools.monitoring;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicLong;
import org.apache.commons.logging.Log;
import com.openexchange.log.LogFactory;
import com.openexchange.office.tools.DocumentType;

/**
 * {@link Statistics
 *
 * @author <a href="mailto:kai.ahrens@open-xchange.com">Kai Ahrens</a>
 */
/**
 * {@link Statistics}
 *
 * @author <a href="mailto:kai.ahrens@open-xchange.com">Kai Ahrens</a>
 */
public class Statistics {

    // set period of reset task to 5min, which equals to the munin standard period
    private final long RESET_TIMER_PERIOD_MILLISECONDS = 5 * 60 * 1000;

    // set ahead time to periodically retrieve results this time
    // span before the next external and periodic retrieval
    private final long RESET_TIMER_AHEAD_TIME = 30 * 1000;

    /**
     * Logger
     */
    protected static final Log LOG = LogFactory.getLog(Statistics.class);

    /**
     * Initializes a new {@link Statistics}.
     */
    public Statistics() {
        super();

        if (null != m_statistics) {
            LOG.warn("Statistics object has already been initialized, please check callers.");
        } else {
            m_statistics = this;
        }

        // initialize members

        for (DocumentType documentType : DocumentType.values()) {
            // correctly initialize all maps and sub maps
            m_documentsCreated.put(documentType, new AtomicLong());
            m_documentsErrors.put(documentType, new AtomicLong());

            // OpenType
            final EnumMap<OpenType, AtomicLong> openMap = new EnumMap<OpenType, AtomicLong>(OpenType.class);
            m_documentsOpened.put(documentType, openMap);

            for (OpenType curOpenType : OpenType.values()) {
                openMap.put(curOpenType, new AtomicLong());
            }

            // OperationsType
            final EnumMap<OperationsType, AtomicLong> operationsMap = new EnumMap<OperationsType, AtomicLong>(OperationsType.class);
            m_documentsOperations.put(documentType, operationsMap);

            for (OperationsType curOperationsType : OperationsType.values()) {
                operationsMap.put(curOperationsType, new AtomicLong());
            }

            // SaveType
            final EnumMap<SaveType, AtomicLong> saveMap = new EnumMap<SaveType, AtomicLong>(SaveType.class);
            m_documentsSaved.put(documentType, saveMap);

            for (SaveType curSaveType : SaveType.values()) {
                saveMap.put(curSaveType, new AtomicLong());
            }

            // CloseType
            final EnumMap<CloseType, AtomicLong> closeMap = new EnumMap<CloseType, AtomicLong>(CloseType.class);
            m_documentsClosed.put(documentType, closeMap);

            for (CloseType curCloseType : CloseType.values()) {
                closeMap.put(curCloseType, new AtomicLong());
            }

            // initialize periodic members
            m_documentsSizes.put(documentType, new ArrayList<AtomicLong>());
            m_documentsSizesResult.put(documentType, new AtomicLong());
        }

        // CalcEngine
        for (CalcEngineEventType curCalcEngineEventType : CalcEngineEventType.values()) {
            m_calcEngineEvents.put(curCalcEngineEventType, new AtomicLong());
        }
    }

    /**
     * @param event
     */
    static public synchronized void handleDocumentEvent(DocumentEvent event) {
        if (null != m_statistics) {
            m_statistics.implHandleDocumentEvent(event);
        } else {
            LOG.warn("Statistics object has not been initialized yet, please check proper initialization.");
        }
    }

    /**
     * @return The 'real' implementation object
     */
    static public Statistics get() {
        return m_statistics;
    }

    /**
     * @return
     */
    public synchronized long getDocumentsCreated(com.openexchange.office.tools.DocumentType documentType) {
        long ret = 0;

        if (null == documentType) {
            for (DocumentType curDocumentType : DocumentType.values()) {
                ret += m_documentsCreated.get(curDocumentType).get();
            }
        } else {
            ret = m_documentsCreated.get(documentType).get();
        }

        return ret;
    }

    /**
     * @return
     */
    public synchronized long getDocumentsOpened(DocumentType documentType, OpenType openType) {
        return getDocumentsCountValue(documentType, m_documentsOpened, openType);
    }

    /**
     * @return
     */
    public synchronized long getDocumentsOperations(DocumentType documentType, OperationsType operationType) {
        return getDocumentsCountValue(documentType, m_documentsOperations, operationType);
    }

    /**
     * @return
     */
    public synchronized long getDocumentsSaved(DocumentType documentType, SaveType saveType) {
        return getDocumentsCountValue(documentType, m_documentsSaved, saveType);
    }

    /**
     * @return
     */
    public synchronized long getDocumentsClosed(DocumentType documentType, CloseType closeType) {
        return getDocumentsCountValue(documentType, m_documentsClosed, closeType);
    }

    /**
     * @return
     */
    public synchronized long getDocumentsSizeMedian(DocumentType documentType) {
        initPeriodicResetTimer();

        return ((null == documentType) ? m_documentsSizesResultTotal.get() : m_documentsSizesResult.get(documentType).get());
    }

    /**
     * @return
     */
    public synchronized long getCalcEngineData(CalcEngineEventType calcEngineEventType) {
        return m_calcEngineEvents.get(calcEngineEventType).get();
    }

    /**
     *
     */
    public synchronized long getCalcEngineMedianJobTime() {
        return m_calcEngineJobTimesResult.get();
    }

    // - Implementation --------------------------------------------------------

    /**
     * @param event
     */
    protected void implHandleDocumentEvent(DocumentEvent event) {
        if (null != event) {
            final DocumentType documentType = event.getDocumentType();

            m_eventCount.incrementAndGet();

            if (null != documentType) {
                switch (event.getEventType()) {
                case CREATE: {
                    m_documentsCreated.get(documentType).incrementAndGet();
                    break;
                }

                case OPEN: {
                    if (event instanceof OpenEvent) {
                        final OpenEvent openEvent = (OpenEvent) event;

                        m_documentsOpened.get(documentType).get(openEvent.getOpenType()).incrementAndGet();
                        m_documentsSizes.get(documentType).add(new AtomicLong(openEvent.getDocumentSize()));
                    }

                    break;
                }

                case OPERATIONS: {
                    if (event instanceof OperationsEvent) {
                        final OperationsEvent operationsEvent = (OperationsEvent) event;

                        m_documentsOperations.get(documentType).get(operationsEvent.getOperationType()).addAndGet(
                            operationsEvent.getActionCount());
                    }
                    break;
                }

                case SAVE: {
                    if (event instanceof SaveEvent) {
                        final SaveEvent saveEvent = (SaveEvent) event;

                        m_documentsSaved.get(documentType).get(saveEvent.getSaveType()).incrementAndGet();
                    }
                    break;
                }

                case CLOSE: {
                    if (event instanceof CloseEvent) {
                        final CloseEvent closeEvent = (CloseEvent) event;

                        m_documentsClosed.get(documentType).get(closeEvent.getCloseType()).incrementAndGet();
                    }
                    break;
                }

                case ERROR: {
                    m_documentsErrors.get(documentType).incrementAndGet();
                    break;
                }

                case CALCENGINE: {
                    if (event instanceof CalcEngineEvent) {
                        final CalcEngineEvent calcEngineEvent = (CalcEngineEvent) event;
                        final CalcEngineEventType calcEngineEventType = calcEngineEvent.getCalcEngineEventType();

                        switch (calcEngineEventType) {
                        case DOC_RELOADED:
                        case JOB_ERROR:
                        case JOB_PROCESSED:
                        case DOC_EXIT:
                        case DOC_ERROR:
                        case DOC_INIT: {
                            m_calcEngineEvents.get(calcEngineEventType).incrementAndGet();

                            if (calcEngineEvent.hasCalcEngineEventDuration()) {
                                m_calcEngineJobTimes.add(new AtomicLong(calcEngineEvent.getCalcEngineEventDuration()));
                            }

                            break;
                        }

                        case NONE:
                        default: {
                            break;
                        }
                        }
                    }

                    break;
                }

                case NONE:
                default: {
                    break;
                }
                }
            }
        }
    }

    /**
    *
    */
    protected void initPeriodicResetTimer() {
        if (null == m_resetTask) {
            // the first values are retrieved => call reset period to retrieve latest results
            resetMeasurementPeriod();

            // the periodically called reset task
            m_resetTask = new TimerTask() {

                /*
                 * (non-Javadoc)
                 * @see com.openexchange.documentconverter.impl.Statistics.ResetTask#run()
                 */
                @Override
                public void run() {
                    resetMeasurementPeriod();
                }
            };

            // start the timer to periodically retrieve further result sets
            // a short amount of time before the next results will be requested
            // with a fixed period from external tools, e.g. 300s with munin
            m_resetTimer.scheduleAtFixedRate(
                m_resetTask,
                RESET_TIMER_PERIOD_MILLISECONDS - RESET_TIMER_AHEAD_TIME,
                RESET_TIMER_PERIOD_MILLISECONDS);
        }
    }

    /**
    *
    */
    protected synchronized void resetMeasurementPeriod() {
        m_documentsSizesResultTotal.set(getDocumentsMedianSizeValue(null));

        for (DocumentType documentType : DocumentType.values()) {
            // save documents sizes results and reset working Map
            m_documentsSizesResult.put(documentType, new AtomicLong(getDocumentsMedianSizeValue(documentType)));
            m_documentsSizes.get(documentType).clear();
        }


        m_calcEngineJobTimesResult.set(getMedianValue(m_calcEngineJobTimes));
        m_calcEngineJobTimes.clear();
    }

    /**
     * @param documentType
     * @param documentsMap
     * @param opType
     * @return
     */
    static protected long getDocumentsCountValue(DocumentType documentType, Map<DocumentType, Map<? extends OpType, AtomicLong>> documentsMap, OpType opType) {
        long ret = 0;

        if ((null != documentType) && (null != documentsMap)) {
            final Map<? extends OpType, AtomicLong> opTypeMap = documentsMap.get(documentType);

            if (null == opType) {
                final Iterator<? extends OpType> keyIter = opTypeMap.keySet().iterator();

                while (keyIter.hasNext()) {
                    ret += opTypeMap.get(keyIter.next()).get();
                }
            } else {
                ret = opTypeMap.get(opType).get();
            }
        }

        return ret;
    }

    /*
     * @param sizesMap
     * @param documentType
     * @return
     */
    protected long getDocumentsMedianSizeValue(DocumentType documentType) {
        ArrayList<AtomicLong> workList = null;

        if (null == documentType) {
            workList = new ArrayList<AtomicLong>();

            for (DocumentType curDocumentType : DocumentType.values()) {
                workList.addAll(m_documentsSizes.get(curDocumentType));
            }
        } else {
            workList = m_documentsSizes.get(documentType);
        }

        return getMedianValue(workList);
    }

    /**
     * @param workList
     * @return
     */
    protected static long getMedianValue(ArrayList<AtomicLong> workList) {
        double ret = 0.0;

        if (null != workList) {
            final long[] longArray = new long[workList.size()];

            if (longArray.length > 0) {
                final int medianIndex = longArray.length >> 1;
                int i = 0;

                for (AtomicLong curEntry : workList) {
                    longArray[i++] = curEntry.get();
                }

                Arrays.sort(longArray);

                // calculate average of left and right values wrt. the median position for even size lists
                ret = (0 == (longArray.length % 2)) ? ((longArray[medianIndex - 1] + longArray[medianIndex]) * 0.5) : longArray[medianIndex];
            }
        }

        return Math.round(ret);
    }


    // - Members ---------------------------------------------------------------

    static Statistics m_statistics = null;

    private Timer m_resetTimer = new Timer("Office statistics", true);

    private TimerTask m_resetTask = null;

    private AtomicLong m_eventCount = new AtomicLong();

    private Map<DocumentType, AtomicLong> m_documentsCreated = new EnumMap<DocumentType, AtomicLong>(DocumentType.class);

    private Map<DocumentType, AtomicLong> m_documentsErrors = new EnumMap<DocumentType, AtomicLong>(DocumentType.class);

    private Map<DocumentType, Map<? extends OpType, AtomicLong>> m_documentsOpened = new EnumMap<DocumentType, Map<? extends OpType, AtomicLong>>(
        DocumentType.class);

    private Map<DocumentType, Map<? extends OpType, AtomicLong>> m_documentsSaved = new EnumMap<DocumentType, Map<? extends OpType, AtomicLong>>(
        DocumentType.class);

    private Map<DocumentType, Map<? extends OpType, AtomicLong>> m_documentsClosed = new EnumMap<DocumentType, Map<? extends OpType, AtomicLong>>(
        DocumentType.class);

    private Map<DocumentType, Map<? extends OpType, AtomicLong>> m_documentsOperations = new EnumMap<DocumentType, Map<? extends OpType, AtomicLong>>(
        DocumentType.class);

    private Map<DocumentType, ArrayList<AtomicLong>> m_documentsSizes = new EnumMap<DocumentType, ArrayList<AtomicLong>>(DocumentType.class);

    private Map<DocumentType, AtomicLong> m_documentsSizesResult = new EnumMap<DocumentType, AtomicLong>(DocumentType.class);

    private AtomicLong m_documentsSizesResultTotal = new AtomicLong();

    private Map<CalcEngineEventType, AtomicLong> m_calcEngineEvents = new EnumMap<CalcEngineEventType, AtomicLong>(CalcEngineEventType.class);

    private ArrayList<AtomicLong> m_calcEngineJobTimes = new ArrayList<AtomicLong>();

    private AtomicLong m_calcEngineJobTimesResult = new AtomicLong();
}
